#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include "linklist.h"
#include "state.h"
#include "hashmap.h"
#include "clock.h"

bool eq_v3(void* a,void* b){
    struct SizedSubmask* mask_a=a;
    struct SizedSubmask* mask_b=b;
    return mask_a->mask.raw==mask_b->mask.raw;
}

struct Node* xor_remove_match(struct Node* head,void* target,bool (*eq)(void *,void *)){
    struct Node* previous=NULL;
    struct Node* current=head;
    while(current!=NULL){
        struct Node* next=xor_next(previous,current);
        if(eq(current->val,target)){
            free(current->val);
            xor_remove(previous,current);
            break;
        }
        previous=current;
        current=next;
    }
    if(previous==NULL) return xor_next(previous,current);
    return NULL;
}

bool xor_contain(struct Node* head,void* target,bool (*eq)(void *,void *)){
    struct Node* previous=NULL;
    struct Node* current=head;
    while(current!=NULL){
        struct Node* next=xor_next(previous,current);
        if(eq(current->val,target)){
            return false;
        }
        previous=current;
        current=next;
    }
    return true;
}

void prefix_insert(struct State* state){
    FILE * table = fopen("inserted_prefixes.txt","r");

    clock_start

    while(true){
        char buffer[20]={'\0'};
        fgets(buffer,20,table);
        if(*buffer=='\0') break;

        struct SizedSubmask* mask=malloc(sizeof(struct SizedSubmask));
        parse_submask(buffer,mask);
        struct SizedSubmask* ip=clone_submask(mask);

        record_start

        if(reduce_submask(mask)){
            struct Node* ll=hash_get(state->hashmap,mask);
            if(ll==NULL){
                struct Node* ll=xor_new(ip);
                hash_insert(state->hashmap,mask,ll);
            }else{
                xor_insert_front(ll,ip);
                free(mask);
            }
        }else{
            free(ip);
            if(state->head==NULL)state->head=xor_new(mask);
            else state->head=xor_insert_mid(NULL,state->head,mask);
        }

        record_end("prefix_insert.csv")

        freq++;
    }

    clock_end
}

void prefix_delete(struct State* state){
    FILE * table = fopen("deleted_prefixes.txt","r");

    clock_start

    while(true){
        char buffer[20]={'\0'};
        fgets(buffer,20,table);
        if(*buffer=='\0') break;


        struct SizedSubmask* mask=malloc(sizeof(struct SizedSubmask));
        parse_submask(buffer,mask);
        struct SizedSubmask* ip=clone_submask(mask);

        record_start

        if(reduce_submask(mask)){
            struct Node* ll=hash_pop(state->hashmap,mask);
            if(ll==NULL) {
                log_warn("removal fail, cannot find index!");
                goto drop;
            }
            struct Node* new_head=xor_remove_match(ll,ip,eq_v3);
            if(new_head==NULL) hash_insert(state->hashmap,mask,ll);
            else hash_insert(state->hashmap,mask,new_head);
        }else{
            struct Node* ll=state->head;
            if(ll==NULL) {
                log_warn("removal fail, cannot find index!");
                goto drop;
            }
            struct Node* new_head=xor_remove_match(ll,ip,eq_v3);
            if(new_head!=NULL) state->head=new_head;
        }
        freq++;

        record_end("prefix_delete.csv")

        drop:
        free(ip);
        free(mask);
    }

    clock_end
}

void search(struct State* state){
    FILE * table = fopen("trace_file.txt","r");

    clock_start

    while(true){
        char buffer[20]={'\0'};
        fgets(buffer,20,table);
        if(*buffer=='\0') break;

        struct SizedSubmask* mask=malloc(sizeof(struct SizedSubmask));
        parse_submask(buffer,mask);
        struct SizedSubmask* ip=clone_submask(mask);

        record_start

        struct Node* ll;
        if(reduce_submask(mask))ll=hash_get(state->hashmap,mask);
        else ll=state->head;
        
        freq++;

        record_end("search.csv")

        printf("          submask: ");
        display_ip(ip);

        if (xor_contain(ll,ip,eq_v3)) printf(" can be found among entries\n");
        else printf(" cannot be found among entries\n");
            
        free(ip);
        free(mask);
    }

    clock_end
}